@@ -2,7 +2,7 @@
 #define SOCKET_HPP
 
 #include "http.h"
-#include <boost/shared_ptr.hpp>
+#include <memory>
 #include "libuv/include/uv.h"
 
 class HttpRequest;
@@ -11,18 +11,18 @@ class WebApplication;
 class Socket {
 public:
   VariantHandle handle;
-  boost::shared_ptr<WebApplication> pWebApplication;
+  std::shared_ptr<WebApplication> pWebApplication;
   CallbackQueue* background_queue;
-  std::vector<boost::shared_ptr<HttpRequest>> connections;
+  std::vector<std::shared_ptr<HttpRequest> > connections;
 
-  Socket(boost::shared_ptr<WebApplication> pWebApplication,
+  Socket(std::shared_ptr<WebApplication> pWebApplication,
          CallbackQueue* background_queue)
     : pWebApplication(pWebApplication), background_queue(background_queue)
   {
   }
 
-  void addConnection(boost::shared_ptr<HttpRequest> request);
-  void removeConnection(boost::shared_ptr<HttpRequest> request);
+  void addConnection(std::shared_ptr<HttpRequest> request);
+  void removeConnection(std::shared_ptr<HttpRequest> request);
   void close();
 
   virtual ~Socket();
